/***************************************************************************
    file	         : kb_display.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	<qpainter.h>
#include	<qcursor.h>

#if		__KB_AUTOMOC
#include	"kb_display.moc"
#else
#include	"kb_display.h"
#endif

#include	"kb_formblock.h"

#include	"kb_options.h"
#include	"kb_layout.h"
#include	"kb_dispwidget.h"
#include	"kb_dispscroller.h"
#include	"kb_report.h"
#include	"kb_ascii.h"
#include	"kb_writer.h"
#include	"kb_toolbox.h"
#include	"kb_popupmenu.h"



/*  snapPosition: Snap position to grid					*/
/*  pos		: int		: Position to be snapped		*/
/*  near	: bool		: True to snap to nearest		*/
/*  x		: bool		: True for X, else Y			*/
/*  (returns)	: int		: Position snapped to grid		*/

int	snapPosition
	(	int	pos,
		bool	bNear,
		bool	x
	)
{
	uint	step	;
	uint	delta	;

	/* Get the appropriate grid step. If zero (only if the user	*/
	/* edits the configuration file) or one then the position is	*/
	/* left unchanged (ie., treat zero as one).			*/
	step	= x ? KBOptions::getGridXStep() : KBOptions::getGridYStep() ;
	if (step <= 1) return pos ;

	/* To avoid round-towards-zero problems (ie, to get round-down	*/
	/* behaviour) we use an offset. THis must be a multiple of the	*/
	/* grid step and larger than any reasonable coordinate.		*/
	delta	= step	;
	while (delta < 100000) delta *= 100 ;

	/* If snapping to nearest rather than left/up, adjust the	*/
	/* position by half the grid step.				*/
	if (bNear) pos += step / 2 ;

	/* Rock and roll. Remember, sex-and-drugs-and-rock-and-roll are	*/
	/* only substitutes for caving.					*/
	return	((pos + delta) / step) * step - delta ;
}

/*  snapRect	: Snap rectangle to nearest grid points			*/
/*  rect	: QRect &	: Area to snap				*/
/*  (returns)	: void		:					*/

void	snapRect
	(	QRect		&rect
	)
{
	int	x	= rect.x()	;
	int	y	= rect.y()	;
	int	w	= rect.width ()	;
	int	h	= rect.height()	;

	w	= snapPosition (x + w, true, true ) ;
	h	= snapPosition (y + h, true, false) ;
	x	= snapPosition (x,     true, true ) ;
	y	= snapPosition (y,     true, false) ;
	w	= w - x ;
	h	= h - y ;

	rect	= QRect (x, y, w, h) ;
}


/*  ------------------------------------------------------------------  */



/*  KBDisplay								*/
/*  KBDisplay	: Constructor for root display window			*/
/*  parent	: QWidget *	: Parent widget				*/
/*  object	: KBObject	: Owning object				*/
/*  showBar	: uint		: Show row scroller bar/mini-nav	*/
/*  stretch	: bool		: Form in stretch mode			*/
/*  showRuler	: bool		: Show design ruler			*/
/*  (returns)	: KBDisplay	:					*/

KBDisplay::KBDisplay
	(	QWidget		*parent,
		KBObject	*object,
		uint		showBar,
		bool		stretch,
		bool		showRuler
	)
	:
	m_parent	(0),
	m_object	(object)
{
	if (m_object->isDynamic())
	{
		m_scroller = 0 ;
		m_widget   = new KBDispWidget     (parent, this, showBar) ;
		m_canvas   = m_widget->getCanvas  () ;
	}
	else
	{
		m_scroller = new KBDispScroller   (parent, this, showBar, stretch, showRuler) ;
		m_widget   = 0	 ;
		m_canvas   = m_scroller->viewport () ;
	}

	m_marking  = false ; 
	m_shown	   = false ;
	m_lMargin  = 0	   ;
	m_rMargin  = 0	   ;
	m_tMargin  = 0	   ;
	m_bMargin  = 0	   ;
}



/*  KBDisplay								*/
/*  KBDisplay	: Constructor for display window			*/
/*  parent	: KBDisplay *	: Parent display			*/
/*  object	: KBObject	: Owning object				*/
/*  showBar	: uint		: Show row scroller bar/mini-nav	*/
/*  (returns)	: KNDisplay	:					*/

KBDisplay::KBDisplay
	(	KBDisplay	*parent,
		KBObject	*object,
		uint		showBar
	)
	:
	m_parent	(parent),
	m_object	(object)
{
	m_scroller = 0	   ;
	m_widget   = new KBDispWidget
		       (	m_parent->getDisplayWidget(),
				this,
				showBar
		       )	;
	m_canvas   = m_widget	;

	m_parent->addChild
	(	m_widget,
		m_object->geometry(),
		m_object->attrGeom().alignment()
	) ;

	m_shown	   = false ;
	m_marking  = false ;
	m_lMargin  = 0	   ;
	m_rMargin  = 0	   ;
	m_tMargin  = 0	   ;
	m_bMargin  = 0	   ;
}

/*  KBDisplay								*/
/*  KBDisplay	: Destructor for display window				*/
/*  (returns)	:		:					*/

KBDisplay::~KBDisplay ()
{
	if 	(m_scroller) delete m_scroller ;
	else if (m_widget  ) delete m_widget   ;
}


QSize	KBDisplay::getTopSize ()
{
	return	m_scroller ?
			m_scroller->getTopSize()    :
			m_widget  ->getTopSize()    ;
}

QRect	KBDisplay::getTopRect ()
{
	if (m_scroller)
		return	QRect(QPoint(0,0), m_scroller->getTopSize()) ;

	QRect	r = m_widget->geometry() ;
	if (m_parent != 0)
		m_parent->cvtViewToCtrl (r) ;
	return	r ;
}


void	KBDisplay::resizeContents
	(	const QSize		&size
	)
{
	if (m_scroller)
		m_scroller->resizeContents (size) ;
	else	m_widget  ->resizeContents (size) ;
}

void	KBDisplay::setRowColSetup
	(	const KBGridSetupList	&rowSetup,
		const KBGridSetupList	&colSetup
	)
{
	if (m_widget)
		m_widget->setRowColSetup (rowSetup, colSetup) ;
}


#if	! __KB_RUNTIME

/*  KBDisplay								*/
/*  drawDisplay	: Handle painting					*/
/*  p		: QPainter *	: Painter				*/
/*  clip	: QRect		: Area to paint				*/
/*  (returns)	: void		:					*/

void	KBDisplay::drawDisplay
	(	QPainter *p,
		QRect	
	)
{
	KBReport *report = m_object->isReport() ;
	double	  ppm	 = pixelsPerMM () ;

	p->setPen(Qt::black) ;

	if (report != 0)
	{
		report->margins (m_lMargin, m_rMargin, m_tMargin, m_bMargin) ;

		m_lMargin = (int)(m_lMargin * ppm) ;
		m_rMargin = (int)(m_rMargin * ppm) ;
		m_tMargin = (int)(m_tMargin * ppm) ;
		m_bMargin = (int)(m_bMargin * ppm) ;
	}

	if (m_marking)
	{
		int	x = m_markX + m_lMargin ;
		int	y = m_markY + m_tMargin ;

		m_rubber  = QRect() ;
		p->save	  () ;
		moveRubberRect (p, QRect(x, y, m_markW, m_markH)) ;
		p->restore() ;
	}

	int	gx	= (int)KBOptions::getGridXStep() ;
	int	gy	= (int)KBOptions::getGridXStep() ;

	QSize	size	= getTopSize  () ;
	int	dw	= size.width  () + m_lMargin ;
	int	dh	= size.height () + m_tMargin ;


//	fprintf
//	(	stderr,
//		"KBDisplay::drawDisplay: %s->(%d,%d)\n",
//		_TEXT(size),
//		dw,
//		dh
//	)	;

	if (gx <= 0) gx = 10 ;
	if (gy <= 0) gy = 10 ;

	for (int px = m_lMargin ; px <= dw -1 ; px += gx)
		for (int py = m_tMargin ; py <= dh - 1 ; py += gy)
			p->drawPoint (px, py) ;


	QRect	newArea	= QRect(m_lMargin, m_tMargin, dw - m_lMargin, dh - m_tMargin) ;
	p->drawRect (newArea) ;

	if (m_scroller)
	{
static QRect m_oldArea ;

		QRect	maxArea	= m_oldArea.unite (newArea) ;
		QBrush bg	(p->backgroundColor()) ;

		if (maxArea.width () > newArea.width ())
			p->fillRect (newArea.width() + m_lMargin,  m_tMargin, maxArea.width()  - newArea.width(),  maxArea.height(), bg) ;
		if (maxArea.height() > newArea.height())
			p->fillRect (m_lMargin, newArea.height() + m_tMargin, maxArea.width(),  maxArea.height() - newArea.height(), bg) ;
m_oldArea = newArea ;
	}

	if (report != 0)
	{
		int	width	;
		int	height	;

		getPixelPageSize("A4", width, height) ;

		QBrush	brush	(QColor(0xe0, 0xe0, 0xe0)) ;

		p->fillRect (0, 0, width, m_tMargin, brush) ;
		p->fillRect (0, height - m_bMargin, width, m_bMargin, brush) ;
		p->fillRect (0, m_tMargin, m_lMargin, height - m_tMargin - m_bMargin, brush) ;
		p->fillRect (width - m_rMargin, m_tMargin, m_rMargin, height - m_tMargin - m_bMargin, brush) ;

		p->drawLine (m_lMargin, 0, m_lMargin, height) ;
		p->drawLine (width - m_rMargin,  0, width - m_rMargin, height) ;
		p->drawLine (width,   0, width,   height) ;

		p->drawLine (0, m_tMargin, width,   m_tMargin) ;
		p->drawLine (0, height - m_bMargin, width, height - m_bMargin) ;
		p->drawLine (0, height,  width,   height) ;
	}
}

/*  KBDisplay								*/
/*  contextEvent: Handler for mouse context menu events			*/
/*  cell	: QRect		: Cell for dynamic geometry		*/
/*  (returns)	: bool		: Event consumed			*/

bool	KBDisplay::contextEvent
	(	QRect		cell
	)
{
	if (m_object->showingDesign ())
	{
		KBPopupMenu *popup = m_object->designPopup(0, cell) ;
		popup->exec (QCursor::pos()) ;
		delete	popup	;
		return	true	;
	}

	return	false	;
}

#endif

/*  KBDisplay								*/
/*  getDisplayWidget							*/
/*		: Get widget used for display				*/
/*  (returns)	: QWidget *	: Display widget			*/

QWidget	*KBDisplay::getDisplayWidget ()
{
	return	m_canvas	;
}

/*  KBDisplay								*/
/*  getTopWidget: Get top-most widget					*/
/*  (returns)	: QWidget *	: Top-most widget			*/

QWidget	*KBDisplay::getTopWidget ()
{
	return	m_scroller ? (QWidget *)m_scroller : (QWidget *)m_widget ;
}

/*  KBDisplay								*/
/*  addChild	: Add child widget					*/
/*  rect	: const QRect &	: Position				*/
/*  (returns)	: void		:					*/

void	KBDisplay::addChild
	(	QWidget			*child,
		const QRect		&rect,
		Qt::AlignmentFlags	align
	)
{
	if (m_scroller)
		m_scroller->addChild  (child, rect, align) ;
	else	m_widget  ->addChild  (child, rect, align) ;
}

/*  KBDisplay								*/
/*  moveChild	: Move child widget					*/
/*  rect	: const QRect &	: Position				*/
/*  (returns)	: void		:					*/

void	KBDisplay::moveChild
	(	QWidget			*child,
		const QRect		&rect,
		Qt::AlignmentFlags	align
	)
{
	if (m_scroller)
		m_scroller->moveChild (child, rect, align) ;
	else	m_widget  ->moveChild (child, rect, align) ;
}

/*  KBDisplay								*/
/*  setGeometry	: Set display geometry					*/
/*  rect	: const QRect &	: Position				*/
/*  (returns)	: void		:					*/

void	KBDisplay::setGeometry
	(	const QRect	&rect
	)
{
//	fprintf
//	(	stderr,
//		"KBDisplay::setGeometry: %s s=%p w=%p\n",
//		_TEXT(rect),
//		(void *)m_scroller,
//		(void *)m_widget
//	)	;

	if (m_scroller)
	{
		m_scroller->setTopSize (rect.size()) ;
		return	;
	}

	if (m_widget)
	{
		KBDisplay *pp = m_parent      ;
		uint	  ty  = m_widget->y() ;

		if (pp == 0)
		{
			m_widget->setTopSize (rect.size()) ;
			return	;
		}

		pp->moveChild  (m_widget, rect, Qt::AlignAuto) ;

		while (pp->getParent() != 0)
		{
			ty += pp->getTopWidget()->y() ;
			pp  = pp->getParent   ()      ;
		}

		if (pp->m_scroller)
			pp->m_scroller->moveTags (m_widget, ty) ;
	}
}

/*  KBDisplay								*/
/*  resizeEvent	: Pass size change up to owning object			*/
/*  (returns)	: void		:					*/

void	KBDisplay::resizeEvent ()
{
	m_object->resizeEvent () ;
}


/*  KBDisplay								*/
/*  width	: Get display width					*/
/*  (returns)	: int		: Width					*/

int	KBDisplay::width ()
{
	/* This is the width of the display widget if this is not a	*/
	/* scroller, or of the scroller if there is one.		*/
	return	m_scroller ? m_scroller->width() : m_widget->width() ;
}

/*  KBDisplay								*/
/*  height	: Get display height					*/
/*  (returns)	: int		: Height					*/

int	KBDisplay::height ()
{
	/* See width comments above ...					*/
	return	m_scroller ? m_scroller->height() : m_widget->height() ;
}

/*  KBDisplay								*/
/*  setPalette	: Set pallette						*/
/*  pal		: const QPalette *: Palette in question			*/
/*  (returns)	:		  : void				*/

void	KBDisplay::setPalette
	(	const QPalette *pal
	)
{
	getDisplayWidget()->setPalette    (*pal) ;
#if	QT_VERSION >= 300
	getDisplayWidget()->setEraseColor (pal->active().background()) ;
#endif
}

/*  KBDisplay								*/
/*  setFont	: Set font						*/
/*  font	: const QFont *	: Font in question			*/
/*  (returns)	:		: void					*/

void	KBDisplay::setFont
	(	const QFont *font
	)
{
	getDisplayWidget()->setFont (*font) ;
}

/*  KBDisplay								*/
/*  getPalette	: Get pallette						*/
/*  (returns)	: const QPalette: The palette				*/

const	QPalette KBDisplay::palette ()
{
	return	getDisplayWidget()->palette () ;
}

/*  KBDisplay								*/
/*  getFont	: Get font						*/
/*  (returns)	: const QFont	: The font				*/

const	QFont	KBDisplay::font ()
{
	return	getDisplayWidget()->font () ;
}

/*  KBDisplay								*/
/*  show	: Show widget						*/
/*  (returns)	: void		:					*/

void	KBDisplay::show ()
{
	if (!m_shown)
	{
		if (m_scroller)
			m_scroller->show () ;
		else	m_widget  ->show () ;

		m_shown	= true	;
	}
}

/*  KBDisplay								*/
/*  displayIsGone: Notification that display widget is gone		*/
/*  (returns)	 : void		:					*/

void	KBDisplay::displayIsGone ()
{
	/* This notification is used to simplify the problem of		*/
	/* ensuring that all widgets are cleaned up, however they are	*/
	/* destroyed, without having lots of order-of-deletion problems	*/ 
	m_scroller = 0 ;
	m_widget   = 0 ;
	m_canvas   = 0 ;
}

/*  KBDisplay								*/
/*  effectiveSize: Get effective display size				*/
/*  (returns)	 : QSize	: Effective size			*/

QSize	KBDisplay::effectiveSize ()
{
	if (m_scroller)
		return	m_scroller->effectiveSize () ;

	if (m_widget  )
		return	m_widget  ->effectiveSize () ;

	return	QSize(0,0) ;
}

/*  KBDisplay								*/
/*  makeVisible	: Try to ensure that area is visible			*/
/*  rCtrl	: const QRect &		: Area in question		*/
/*  reason	: QFocusEvent::Reason	: Focus reason			*/
/*  (returns)	: void		:					*/

void	KBDisplay::makeVisible
	(	const QRect		&rCtrl,
		QFocusEvent::Reason	reason
	)
{
	/* If the display is not a scroller then there is nothing to do	*/
	/* For a scroller, try to move the specified area into view,	*/
	/* plus some space round it.					*/
	if (m_scroller && m_scroller->isVisible())
	{
//		fprintf
//		(	stderr,
//			"KBDisplay::makeVisible: (%d,%d)(%d)\n",
//			rCtrl.x(),
//			rCtrl.y(),
//			(int)reason
//		)	;

		m_scroller->makeVisible (rCtrl, reason) ;
	}
}

/*  KBDisplay								*/
/*  cvtViewToCtrl: Convert geometry from viewport to contents		*/
/*  cRect	 : QRect &	: Position-and-size			*/
/*  (returns)	 : void		:					*/

void	KBDisplay::cvtViewToCtrl
	(	QRect	&cRect
	)
{
	/* If this is a scroll view then map the control geometry back	*/
	/* to contents coordinates; if not then it is OK already.	*/
	if (m_scroller)
		cRect	= QRect (m_scroller->viewportToContents (cRect.topLeft    ()),
				 m_scroller->viewportToContents (cRect.bottomRight())) ;
}

/*  KBDisplay								*/
/*  cvtCtrlToView: Convert geometry from contents to viewport		*/
/*  cRect	 : QRect &	: Position-and-size			*/
/*  (returns)	 : void		:					*/

void	KBDisplay::cvtCtrlToView
	(	QRect	&cRect
	)
{
	/* If this is a scroll view then map the control geometry to	*/
	/* viewport coordinates; if not then it is OK already.	*/
	if (m_scroller)
		cRect	= QRect (m_scroller->contentsToViewport (cRect.topLeft    ()),
				 m_scroller->contentsToViewport (cRect.bottomRight())) ;
}

/*  KBDisplay								*/
/*  getScroller	: Get scroller widget if any				*/
/*  (returns)	: QScrollView *	: Scroll view or null			*/

//QScrollView *KBDisplay::getScroller ()
//{
//	return	m_scroller ? m_scroller->getScroller () : 0 ;
//}

/*  KBDisplay								*/
/*  setRowRange	: Set range of rows being displayed			*/
/*  totalRows	: uint		: Total number of rows			*/
/*  extra	: uint		: Extra insertion rows			*/
/*  curQRow	: uint		: Current query row			*/
/*  curDRow	: uint		: Index of topmost displayed row	*/
/*  numRows	: uint		: Number of rows actually visible	*/
/*  (returns)	: void		:					*/

void	KBDisplay::setRowRange
	(	uint	totalRows,
		uint	extra,
		uint	curQRow,
		uint	curDRow,
		uint	numRows
	)
{
	if (m_scroller)
		m_scroller->setRowRange (totalRows, extra, curQRow, curDRow, numRows) ;
	else	m_widget  ->setRowRange (totalRows, extra, curQRow, curDRow, numRows) ;
}

/*  KBDisplay								*/
/*  scrollToRow	: Pass row scroll up to owning form block if any	*/
/*  row		: uint		: Row number				*/
/*  (returns)	: void		:					*/

void	KBDisplay::scrollToRow
	(	uint	row
	)
{
	KBFormBlock *block = m_object->isFormBlock () ;

	if (block == 0) block = m_object->getFormBlock () ;
	if (block != 0) block->scrollToRow (row) ;
}

/*  KBDisplay								*/
/*  scrollBy	: Scroll by number of rows				*/
/*  delta	: int		: Number of rows			*/
/*  (returns)	: void		:					*/

void	KBDisplay::scrollBy
	(	int	delta
	)
{
	KBFormBlock *block = m_object->isFormBlock () ;

	if (block == 0) block = m_object->getFormBlock () ;
	if (block != 0) block->scrollBy (delta) ;
}

/*  KBDisplay								*/
/*  doOperation	: Pass action though to block				*/
/*  action	: KB::Action	: Action in question			*/
/*  qrow	: uint		: Query for for gotorow			*/
/*  (returns)	: void		:					*/

void	KBDisplay::doOperation
	(	KB::Action	action,
		uint		qrow
	)
{
	KBFormBlock *block = m_object->isFormBlock () ;

	if (block == 0) block = m_object->getFormBlock () ;
	if (block != 0) block->doOperation (action, qrow) ;
}

/*  KBDisplay								*/
/*  keyStroke	: Handle keystroke					*/
/*  key		: QKeyEvent *	: Key event				*/
/*  (returns)	: bool		: Event consumed			*/

bool	KBDisplay::keyStroke
	(	QKeyEvent	*key
	)
{
	KBFormBlock *block = m_object->isFormBlock () ;
	return	block == 0 ? false : block->KBNavigator::keyStroke (0, key) ;
}

/*  KBDisplay								*/
/*  showAs	: Set or clear design mode				*/
/*  mode	: KB::ShowAs	: Design mode				*/
/*  (returns)	: void		:					*/

void	KBDisplay::showAs
	(	KB::ShowAs	mode
	)
{
	if (m_scroller)
		m_scroller->showAs (mode) ;
	else	m_widget  ->showAs (mode) ;
}

/*  KBDisplay								*/
/*  setStretchable							*/
/*		: Set or clear stretch mode				*/
/*  stretch	: bool		: Can stretch				*/
/*  (returns)	: void		:					*/

void	KBDisplay::setStretchable
	(	bool	stretch
	)
{
	if (m_scroller)
		m_scroller->setStretchable (stretch) ;
}

/*  KBDisplay								*/
/*  setShowbar	: Set or clear scrollbar mode				*/
/*  showBar	: uint		: Show row scroller bar/mini-nav	*/
/*  (returns)	: void		:					*/

void	KBDisplay::setShowbar
	(	uint	showbar
	)
{
	if (m_scroller)
		m_scroller->setShowbar (showbar) ;
	else	m_widget  ->setShowbar (showbar) ;
}

/*  KBDisplay								*/
/*  setTitle	: Set "group box" title					*/
/*  title	: const QString & : Title text				*/
/*  (returns)	: void		  :					*/

void	KBDisplay::setTitle
	(	const QString	&title
	)
{
	if (m_widget) m_widget->setTitle (title) ;
}

void	KBDisplay::updateDynamic ()
{
	if (m_scroller)
		m_scroller->updateDynamic (m_object->attrGeom()) ;
	else	m_widget  ->updateDynamic (m_object->attrGeom()) ;
}

/*  KBDisplay								*/
/*  setFrame	: Set frame atyle					*/
/*  title	: const QString & : Title text				*/
/*  (returns)	: void		  :					*/

void	KBDisplay::setFrame
	(	int	style,
		int	width
	)
{
//	fprintf
//	(	stderr,
//		"KBDisplay::setFrame: %s: w=%d s=%d\n",
//		m_object->getName().latin1(),
//		width,
//		style
//	)	;

	if (m_widget)
		m_widget->setFrame (style, width) ;
}

/*  KBDisplay								*/
/*  sizeHint	: Get size hint for dynamic displays			*/
/*  (returns)	: QSize		: Base size				*/

QSize	KBDisplay::sizeHint ()
{
	return	m_canvas->sizeHint() ;
}

/*  KBDisplay								*/
/*  morphDestoyed: Catch destuction of morphed control			*/
/*  (returns)	 : void		:					*/

void	KBDisplay::morphDestroyed ()
{
	m_morphList.removeRef ((KBItem *)sender()) ;
}

/*  KBDisplay								*/
/*  addMorph	: Add a morphing control				*/
/*  morph	: KBItem *	: Control in question			*/
/*  (returns)	: void		:					*/

void	KBDisplay::addMorph
	(	KBItem		*morph
	)
{
	m_morphList.append (morph) ;
	connect	(morph, SIGNAL(destroyed()), SLOT(morphDestroyed())) ;
}

/*  KBDisplay								*/
/*  remMorph	: Remove a morphing control				*/
/*  morph	: KBItem *	: Control in question			*/
/*  (returns)	: void		:					*/

void	KBDisplay::remMorph
	(	KBItem		*morph
	)
{
	if (m_morphList.removeRef (morph))
		disconnect (morph, SIGNAL(destroyed()), this, SLOT(morphDestroyed())) ;
}

/*  KBDisplay								*/
/*  updateMorph	: Update a morphing control				*/
/*  morph	: KBItem *	: Control in question			*/
/*  mRect	: const QRect &	: Area occupied by control		*/
/*  (returns)	: void		:					*/

void	KBDisplay::updateMorph
	(	KBItem		*morph,
		const QRect	&mRect
	)
{
	if (m_scroller)
	{	QRect	vRect	= mRect	;
		cvtCtrlToView (vRect)	;
		m_scroller->updateMorph (morph, vRect) ;
	}
	else	m_widget  ->updateMorph	(morph, mRect) ;
}

/*  KBDisplay								*/
/*  updateMorph	: Update a morphing control				*/
/*  morph	: KBItem *	: Control in question			*/
/*  drow	: uint		: Row affected				*/
/*  (returns)	: void		:					*/

void	KBDisplay::updateMorph
	(	KBItem		*morph,
		uint		drow
	)
{
	if (m_scroller)
		m_scroller->updateMorph (morph, drow) ;
	else	m_widget  ->updateMorph (morph, drow) ;
}

void	KBDisplay::repaintMorphs
	(	QPainter	&p,
		const QRect	&rect
	)
{
	LITER
	(	KBItem,
		m_morphList,
		morph,
		morph->repaintMorph (&p, rect) ;
	)

}

/*  KBDisplay								*/
/*  setTagLabel	: Set design mode tag label text			*/
/*  text	: const QString & : Text				*/
/*  (returns)	: void		  :					*/

void	KBDisplay::setTagLabel
	(	const QString	&text
	)
{
	if (m_widget)
		m_widget->setTagLabel (text) ;
}

/*  KBDisplay								*/
/*  setBackgroundPixmap							*/
/*		: Set bgackground pixmap				*/
/*  pixmap	: const QPixmap & : Pixmap				*/
/*  scale	: int		  : Sclaing				*/
/*  (returns)	: void		  :					*/

void	KBDisplay::setBackgroundPixmap
	(	const QPixmap	&pixmap,
		int		scale
	)
{
	if (m_widget)
		m_widget->setBackgroundPixmap (pixmap, scale) ;
}

/*  KBDisplay								*/
/*  mousePressEvent							*/
/*		: Handler for mouse press events			*/
/*  e		: QMouseEvent *	: Mouse event				*/
/*  cell	: QPoint	: Cell for dynamic geometry		*/
/*  (returns)	: bool		: Event consumed			*/

bool	KBDisplay::mousePressEvent
	(	QMouseEvent	*e,
		QPoint		cell
	)
{
	m_rubber = QRect() ;

	if (e->button() != QMouseEvent::LeftButton)
		return	false	;


	LITER
	(	KBItem,
		m_morphList,
		morph,
		if (morph->mouseClickHit (e->pos()))
			return true	;
	)

#if	! __KB_RUNTIME
	if (!m_object->showingDesign ()) return false ;
	if (m_marking) return true ;


	m_left	= 0	 ;
	m_top	= 0	 ;

	/* If the owner is a report then get the left and top margins	*/
	/* and adjust the position accordingly; if not then the		*/
	/* effective margin is zero.					*/
	KBReport *report ;
	if ((report = m_object->isReport()) != 0)
	{
		report->margins (m_left, m_top) ;

		m_left	= (int)(m_left * pixelsPerMM()) ;
		m_top	= (int)(m_top  * pixelsPerMM()) ;
	}

	int x	= e->x() - m_left ;
	int y	= e->y() - m_top  ;

	/* At the top level, check that we are inside the block area.	*/
	/* In either case, set the marking area bounds.			*/
	m_bounds	= getTopSize ()	;
	if ((x >= m_bounds.width()) || (y >= m_bounds.height()))
		return	false	;

	if ((x < 0) || (y < 0))
		return	false	;

	m_marking	= true	;
	m_startMX	= m_markX = x	;
	m_startMY	= m_markY = y	;
	m_startCell	= cell		;
	m_markW		= m_markH = 0	;

	m_canvas->grabMouse ()  ;
	return	true	;
#endif

	return	false	;
}

/*  KBDisplay								*/
/*  mouseMoveEvent							*/
/*		: Handle mouse movement during marking			*/
/*  e		: QMouseEvent *	: Mouse event				*/
/*  (returns)	: bool		: Event consumed			*/

bool	KBDisplay::mouseMoveEvent
	(	QMouseEvent 	*e
	)
{
#if	! __KB_RUNTIME
	if ((e->state() & Qt::LeftButton) == 0) return true ;
	if (!m_marking) return true ;

	int	x = e->x() ;
	int	y = e->y() ;

	/* Zaurus fix: We seem to get a (-2001,-2001) as the button is	*/
	/* released ...							*/
	if ((x < 0) || (y < 0)) return true ;

	/* Adjust by the margins if any, and then bound to the current	*/
	/* bounding rectangle.						*/
	x	-= m_left ;
	y	-= m_top  ;

	if (x >= m_bounds.width ()) x = m_bounds.width () - 1 ;
	if (y >= m_bounds.height()) y = m_bounds.height() - 1 ;

	if (x < 0) x = 0 ;
	if (y < 0) y = 0 ;


	if (x >= m_startMX)
		m_markX = m_startMX, m_markW = x - m_startMX ;
	else	m_markX = x,	     m_markW = m_startMX - x ;
	if (y >= m_startMY)
		m_markY = m_startMY, m_markH = y - m_startMY ;
	else	m_markY = y,	     m_markH = m_startMY - y ;

	QPainter p(m_canvas)  ;
	QRect	 r(m_markX + m_lMargin, m_markY + m_tMargin, m_markW, m_markH) ;
	cvtCtrlToView (r)     ;
	moveRubberRect(&p, r) ;
#endif
	return	true	;
}

/*  KBDisplay								*/
/*  mouseReleaseEvent							*/
/*		: Handle mouse release to end marking			*/
/*  e		: QMouseEvent *	: Mouse event				*/
/*  cell	: QPoint	: Cell for dynamic displaus		*/
/*  (returns)	: bool		: Event consumed			*/

bool	KBDisplay::mouseReleaseEvent
	(	QMouseEvent 	*e,
		QPoint		cell
	)
{
#if	! __KB_RUNTIME

	m_canvas->releaseMouse () ;
	if (!m_marking) return false ;
	m_marking = false    ;


	/* Small movements are ignored and suffice to select the object	*/
	/* in which the click took place.				*/
	if ((m_markW < 5) && (m_markH < 5))
	{
		endRubberRect () ;
		m_object->getLayout()->addSizer (m_object->getSizer(), false) ;
		return	true ;
	}


	/* If marking, then there are two possibilities, marking out to	*/
	/* create a new object, and marking out to select existing	*/
	/* objects.							*/
	int	cellx	= m_startCell.x()  ;
	int	celly	= m_startCell.y()  ;
	int	cellw	= cell.x() - cellx ;
	int	cellh	= cell.y() - celly ;

	if (cellw < 0) { cellx += cellw ; cellw = -cellw ; }
	if (cellh < 0) { celly += cellh ; cellh = -cellh ; }

	cellw	+= 1	;
	cellh	+= 1	;


	/* If the object is a report then adjust for the margins. These	*/
	/* will have been noted when the mouse was pressed.		*/
	if (m_object->isReport())
	{
		cellx	-= m_left ;
		celly	-= m_top  ; 
	}

	QRect	rect	(cellx, celly, cellw, cellh) ;

	/* If the shift key is down we always interpret the range as	*/
	/* identifying objects that it covers.				*/
	if ((e->state() & Qt::ShiftButton) != 0)
	{
		endRubberRect () ;
		m_object->selectOverlap (rect) ;
		return	true ;
	}

	fprintf
	(	stderr,
		"KBDisplay::mouseReleaseEvent: rect=%s\n",
		_TEXT(rect)
	)	;

	/* If the toolbox is in use then check for overlap. On overlap	*/
	/* select as above, otherwise it is a request for a new		*/
	/* object.							*/
	if (KBOptions::getUseToolbox() && KBToolBox::self()->showing())
	{
		if (m_object->checkOverlap (rect))
		{
			endRubberRect () ;
			m_object->selectOverlap (rect) ;
			return	true ;
		}

		m_object->addNewObject (rect) ;
		endRubberRect () ;
		return	true ;
	}

	/* No toolbox, then if there is no overlap, add a new object	*/
	/* for the specified area.					*/
	if (!m_object->checkOverlap (rect))
		m_object->addNewObject (rect) ;

	endRubberRect () ;
#endif
	return	true ;
}

QSize	KBDisplay::showBarUsed ()
{
	if (m_scroller) return m_scroller->showBarUsed () ;
	if (m_widget  ) return m_widget  ->showBarUsed () ;

	return	QSize (0, 0) ;
}

#if	! __KB_RUNTIME

void	KBDisplay::moveRubberRect
	(	QPainter	*p,
		const QRect	&rect
	)
{
	p->setPen     (Qt::white ) ;
	p->setRasterOp(Qt::XorROP) ;

	if (m_rubber.isValid())
		p->drawRect (m_rubber) ;

	m_rubber = rect ;
	p->drawRect (m_rubber) ;
}

void	KBDisplay::endRubberRect ()
{
	if (m_rubber.isValid())
	{
		QPainter p (m_canvas) ;

		p.setPen     (Qt::white ) ;
		p.setRasterOp(Qt::XorROP) ;
		p.drawRect   (m_rubber  ) ;

		m_rubber = QRect() ;
	}
}

#endif


