/***************************************************************************
    file	         : kb_geometry.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#include	<stdio.h>

#include	<qapplication.h>
#include	<qlayout.h>
#include	<qpainter.h>

#include	"kb_object.h"
#include	"kb_display.h"
#include	"kb_docroot.h"
#include	"kb_dialog.h"
#include	"kb_ascii.h"

#if		__KB_AUTOMOC
#include	"kb_geometry.moc"
#else		
#include	"kb_geometry.h"
#endif


#define		__KB_DEBUG_GEOM		(0)

class	KBGridPosition
{
public	:

	QRect			m_rect	;
	Qt::AlignmentFlags	m_align	;

	inline	KBGridPosition
		(	const QRect		&rect,
			Qt::AlignmentFlags	align
		)
		:
		m_rect	(rect),
		m_align	(align)
	{
	}

}	;

class	KBGridLayout : public QGridLayout
{
	KBDisplay	*m_display	;

public	:

	KBGridLayout	(QWidget *, int, int, KBDisplay *) ;

	virtual	void	setGeometry (const QRect &) ;
}	;

KBGridLayout::KBGridLayout
	(	QWidget		*parent,
		int		nRows,
		int		nCols,
		KBDisplay	*display
	)
	:
	QGridLayout (parent, nRows, nCols),
	m_display   (display)
{
}

void	KBGridLayout::setGeometry
	(	const QRect	&r
	)
{
	QSize	sbu	= m_display->showBarUsed() ;
	QGridLayout::setGeometry
	(	QRect
		(	r.x	(),
			r.y	(),
			r.width () - sbu.width (),
			r.height() - sbu.height()
	)	)	;
}


/*  KBGeometry								*/
/*  KBGeometry	: Constructor for geometry management class		*/
/*  widget	: QWidget *	: Widget being managed			*/
/*  display	: KBDisplay *	: Associated display object		*/
/*  (returns)	: KBGeometry	:					*/

KBGeometry::KBGeometry
	(	QWidget			*widget,
		KBDisplay		*display
	)
	:
	m_widget	(widget),
	m_display	(display)
{
	m_widgetMap.setAutoDelete (true) ;

	m_grid		= 0	;
	m_reactivate	= false	;

	const KBAttrGeom &geom = display->getObject()->attrGeom() ;
	if (geom.manage() == KBAttrGeom::MgmtDynamic)
	{
		m_grid	= new KBGridLayout
				(	m_widget,
					geom.numRows(false),
					geom.numCols(false),
					m_display
				)	;

		initGrid (geom)	;
		return	;
	}
}

void	KBGeometry::setRowColSetup
	(	const KBGridSetupList	&rowSetup,
		const KBGridSetupList	&colSetup
	)
{
	for (int row = 0 ; row < m_grid->numRows() ; row += 1)
	{
		int	spacing	= 0 ;
		int	stretch	= 0 ;

		if (row < (int)rowSetup.count())
		{	spacing	= rowSetup[row].m_spacing ;
			stretch = rowSetup[row].m_stretch ;
		}

#if	__KB_DEBUG_GEOM
		fprintf
		(	stderr,
			"KBGeometry::setRowColSetup: row %d: sp=%d,st=%d\n",
			row,
			spacing,
			stretch
		)	;
#endif
#if	QT_VERSION >= 0x030200
		m_grid->setRowSpacing (row, spacing) ;
#endif
		m_grid->setRowStretch (row, stretch) ;
	}

	for (int col = 0 ; col < m_grid->numCols() ; col += 1)
	{
		int	spacing	= 0 ;
		int	stretch	= 0 ;

		if (col < (int)colSetup.count())
		{	spacing	= colSetup[col].m_spacing ;
			stretch = colSetup[col].m_stretch ;
		}

#if	__KB_DEBUG_GEOM
		fprintf
		(	stderr,
			"KBGeometry::setRowColSetup: col %d: sp=%d,st=%d\n",
			col,
			spacing,
			stretch
		)	;
#endif
#if	QT_VERSION >= 0x030200
		m_grid->setColSpacing (col, spacing) ;
#endif
		m_grid->setColStretch (col, stretch) ;
	}

	m_reactivate = true   ;
	m_widget->update   () ;

	m_display->getObject()->getDocRoot()->doLayoutChanged() ;
}

/*  KBGeometry								*/
/*  initGrid	: Common gird initialisation stuff			*/
/*  geom	: KBAttrGeom &	: Geometry inforamtion			*/
/*  (returns)	: void		:					*/

void	KBGeometry::initGrid
	(	const KBAttrGeom	&geom
	)
{
	int	margin	= geom.margin () ;
	int	spacing	= geom.spacing() ;

//	fprintf
//	(	stderr,
//		"KBGeometry::initGrid: %s: m=%d s=%d\n",
//		m_display->getObject()->getName().latin1(),
//		margin,
//		spacing
//	)	;

	m_grid->setMargin  (margin  < 0 ? KBDialog::marginHint () : margin ) ;
	m_grid->setSpacing (spacing < 0 ? KBDialog::spacingHint() : spacing) ;

	setRowColSetup
	(	geom.rowSetup(),
		geom.colSetup()
	)	;
}


/*  KBGeometry								*/
/*  insertIntoGrid							*/
/*		: Insert a widget into the grid				*/
/*  child	: QWidget *	: Widget in question			*/
/*  rect	: const QRect & : Position				*/
/*  (returns)	: void		:					*/

void	KBGeometry::insertIntoGrid
	(	QWidget			*child,
		const QRect		&rect,
		Qt::AlignmentFlags	align
	)
{
#if	__KB_DEBUG_GEOM
	fprintf
	(	stderr,
		"KBGeometry::insertIntoGrid: %s\n",
		_TEXT(rect)
	)	;
#endif
	/* Insert the widget, either multi-cell or single as		*/
	/* appropriate. Could do it all as multi-cell (abliet 1x1) but	*/
	/* maybe this is more efficient.				*/
	if ((rect.width() > 1) || (rect.height() > 1))
		m_grid->addMultiCellWidget
		(	child,
			rect.y(),
			rect.y() + rect.height() - 1,
			rect.x(),
			rect.x() + rect.width () - 1,
			align
		)	;
	else	m_grid->addWidget (child, rect.y(), rect.x(), align) ;

	m_reactivate = true   ;
}

/*  KBGeometry								*/
/*  addChild	: Add a child wodget to the layout			*/
/*  child	: QWidget *	: Child in question			*/
/*  rect	: const QRect &	: Position				*/
/*  (returns)	: void		:					*/

void	KBGeometry::addChild
	(	QWidget			*child,
		const QRect		&rect,
		Qt::AlignmentFlags	align
	)
{
#if	__KB_DEBUG_GEOM
	fprintf
	(	stderr,
		"KBGeometry::addChild: %s\n",
		_TEXT(rect)
	)	;
#endif
	if (m_grid != 0)
	{
		m_widgetMap.remove(child) ;
		m_widgetMap.insert(child, new KBGridPosition (rect, align)) ;
		insertIntoGrid	  (child, rect, align) ;

		m_widget->update  () ;
		m_display->getObject()->getDocRoot()->doLayoutChanged() ;

		connect
		(	child,
			SIGNAL	(destroyed(QObject *)),
			SLOT	(slotWidgetDestroyed(QObject *))
		)	;

		return	;
	}

	child->setGeometry (rect) ;
}

/*  KBGeometry								*/
/*  moveChild	:Move a child in the layout				*/
/*  child	: QWidget *	: Child in question			*/
/*  rect	: const QRect &	: Position				*/
/*  (returns)	: void		:					*/

void	KBGeometry::moveChild
	(	QWidget			*child,
		const QRect		&rect,
		Qt::AlignmentFlags	align
	)
{
#if	__KB_DEBUG_GEOM
	fprintf
	(	stderr,
		"KBGeometry::moveChild: %s\n",
		_TEXT(rect)
	)	;
#endif
	if (m_grid != 0)
	{
		KBGridPosition *at = m_widgetMap.find (child) ;
		if ((at != 0) && (at->m_rect == rect)) return ;

		m_grid->remove    (child) ;
		m_widgetMap.remove(child) ;
		m_widgetMap.insert(child, new KBGridPosition (rect, align)) ;
		insertIntoGrid	  (child, rect, align) ;

		m_widget->update  () ;
		m_display->getObject()->getDocRoot()->doLayoutChanged() ;
		return	;
	}

	child->setGeometry (rect) ;
}

/*  KBGeometry								*/
/*  resize	: Resize geometry					*/
/*  size	: const QSize &	: New size				*/
/*  (returns)	: void		:					*/

void	KBGeometry::resize
	(	const QSize	&size
	)
{
#if	__KB_DEBUG_GEOM
	fprintf
	(	stderr,
		"KBGeometry::resize: [%p] (%d,%d) [%p]\n",
		(void *)this,
		size.width (),
		size.height(),
		m_grid
	)	;
#endif

	if (m_grid == 0)
	{
		const QPtrList<KBNode> &children = m_display->getObject()->getChildren() ;
		TITER
		(	Object,
			children,
			child,

			child->updateDisplay() ;
		)		
	}
	else	m_grid->setGeometry (QRect (0, 0, size.width(), size.height())) ;

	m_display->resizeEvent () ;
}

/*  KBGeometry								*/
/*  isDynamic	: Check whether layout is dynamic or not		*/
/*  (returns)	: bool		: True if dynamic			*/

bool	KBGeometry::isDynamic ()
{
	return	m_grid	!= 0 ;
}

/*  KBGeometry								*/
/*  outlineCells: Draw cell outlines for design mode			*/
/*  paint	: QPainter &	: Painter				*/
/*  (returns)	: void		:					*/

void	KBGeometry::outlineCells
	(	QPainter	&paint
	)
{
	if (m_grid != 0)
	{
		/* Nasty hack: After a change is made in design mode	*/
		/* which changes the grid sizes, we need to call the	*/
		/* QGridLayout::activate() method *before* outlining	*/
		/* the cells, otherwise we still get the old sizes.	*/
		if (m_reactivate)
		{
			qApp  ->processEvents() ;
			m_grid->activate     () ;
			m_reactivate = false	;
		}

		outlineCells (m_grid, paint) ;
	}
}

/*  KBGeometry								*/
/*  getCell	: Get cell which contains specfied coordinate		*/
/*  p		: QPoint	: Coordinates				*/
/*  (returns)	: QPoint	: (row, col)				*/

QPoint	KBGeometry::getCell
	(	QPoint		p
	)
{
	/* If we have a dynamic grid layout then may the (mouse)	*/
	/* coordinates to the underlying cell in the grid.		*/
	if (m_grid != 0)
	{
		int	cx	= 0	;
		int	cy	= 0	;

		while (cx < m_grid->numCols() - 1)
		{	if (p.x() < m_grid->cellGeometry (0, cx + 1).x ()) break ;
			cx += 1	;
		}
		while (cy < m_grid->numRows() - 1)
		{	if (p.y() < m_grid->cellGeometry (cy + 1, 0).y()) break ;
			cy += 1	;
		}

#if	__KB_DEBUG_GEOM
		fprintf
		(	stderr,
			"KBGeometry::getCell: (%d,%d) => (%d,%d)\n",
			p.x(), p.y(),
			cx, cy
		)	;
#endif
		/* Note: The result is returned as (col,row) which is	*/
		/* consistent with (x,y) but is the opposite to the	*/
		/* QGridLayout (row,col) interface.			*/
		return	QPoint(cx, cy) ;
	}

	/* If not dynamic then the position is returned unchanged. In	*/
	/* effect this method is mapping a physical position to a	*/
	/* logical position.						*/
	return	p ;
}

/*  KBGeometry								*/
/*  updateDynamic: Update dynamic layout stuff				*/
/*  geom	 : KBAttrGeom &	: Geometry and other information	*/
/*  (returns)	 : void		:					*/

void	KBGeometry::updateDynamic
	(	const KBAttrGeom	&geom
	)
{
	if (m_grid == 0) return ;

	uint	nRows	= geom.numRows(false) ;
	uint	nCols	= geom.numCols(false) ;

	uint	oRows	= m_grid->numRows() ;
	uint	oCols	= m_grid->numCols() ;

	/* Seems to be no way to change the number of rows and columns,	*/
	/* so if these change, delete the current layout and create a	*/
	/* new one, which we repopulate.				*/
	if ((nRows != oRows) || (nCols != oCols))
	{
#if	__KB_DEBUG_GEOM
		fprintf
		(	stderr,
			"KBGeometry::updateDynamic: (%d,%d) -> (%d,%d)\n",
			oRows, oCols,
			nRows, nCols
		)	;
#endif
		delete	m_grid	;
		m_grid	= new KBGridLayout (m_widget, nRows, nCols, m_display) ;

		for
		(	QPtrDictIterator<KBGridPosition> iter (m_widgetMap) ;
			iter.current() != 0 ;
			++iter
		)
		{
			KBGridPosition *pos = iter.current() ;
			insertIntoGrid ((QWidget *)iter.currentKey(), pos->m_rect, pos->m_align) ;
		}
	}

	initGrid (geom)	    ;
	m_widget->update () ;
	m_display->getObject()->getDocRoot()->doLayoutChanged() ;

}

/*  KBGeometry								*/
/*  slotWidgetDestroyed							*/
/*		:: Catch destruction of child widget			*/
/*  object	: QObject *	: Destroyed object, actually a widget	*/
/*  (returns)	: void		:					*/

void	KBGeometry::slotWidgetDestroyed
	(	QObject		*object
	)
{
#if	__KB_DEBUG_GEOM
	fprintf
	(	stderr,
		"KBGeometry::slotWidgetDestroyed: %p\n",
		(void *)object
	)	;
#endif
	m_widgetMap.remove (object) ;

	QApplication::sendPostedEvents (0, QEvent::ChildRemoved) ;

	m_reactivate = true ;
	m_widget ->update   () ;

	/* Take care: when closing the document, the document root may	*/
	/* no longer exists ...						*/
	KBNode	*root   = m_display->getObject()->getRoot() ;
	if ((root != 0) && (root->isDocRoot() != 0))
		root->isDocRoot()->doLayoutChanged() ;
}

/*  KBGeometry								*/
/*  sizeHint	: Get size hint for this display			*/
/*  (returns)	: QSize		: Size hint				*/

QSize	KBGeometry::sizeHint ()
{
#if	__KB_DEBUG_GEOM
	fprintf
	(	stderr,
		"KBGeometry::sizeHint: %s: %p\n",
		(cchar *)m_display->getObject()->getName(),
		(void  *)m_grid
	)	;
#endif
	if (m_grid != 0)
		return	m_grid->sizeHint() ;

	return	m_display->getObject()->sizeHint() ;
}


/*  ------------------------------------------------------------------	*/

/*  outlineCells: Draw cell outlines for design mode			*/
/*  grid	: QGridLayout *	: Qrid layout				*/
/*  paint	: QPainter &	: Painter				*/
/*  (returns)	: void		:					*/

void	KBGeometry::outlineCells
	(	QGridLayout	*grid,
		QPainter	&paint
	)
{
#if	__KB_DEBUG_GEOM
	fprintf
	(	stderr,
		"KBGeometry::outlineCells: %d,%d\n",
		grid->numCols(),
		grid->numRows()
	)	;
#endif
	paint.setPen (QPen(Qt::DashLine)) ;

	for (int col = 1 ; col < grid->numCols() ; col += 1)
	{
		int	x = grid->cellGeometry(0, col).x() ;
		paint.drawLine (x, 0, x, grid->mainWidget()->height()) ;
	}

	for (int row = 1 ; row < grid->numRows() ; row += 1)
	{
		int	y = grid->cellGeometry(row, 0).y() ;
		paint.drawLine (0, y, grid->mainWidget()->width (), y) ;
	}
}

