/***************************************************************************
    file	         : kb_docroot.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/

#if		__KB_AUTOMOC
#include	"kb_docroot.moc"
#else
#include	"kb_docroot.h"
#endif

#include	"kb_callback.h"
#include	"kb_script.h"
#include	"kb_import.h"
#include	"kb_form.h"
#include	"kb_display.h"
#include	"kb_param.h"
#include	"kb_appptr.h"
#include	"kb_paramsetdlg.h"
#include	"kb_progress.h"
#include	"kb_ascii.h"
#include	"kb_mimefactory.h"


/*  KBDocRoot								*/
/*  KBDoctRoot	: Common document root class constructor		*/
/*  parent	: KBNode *	    : Parent (actually, enclosing) node	*/
/*  children	: QPtrList<KBNode> &: Children of said node		*/
/*  location	: KBLocation &	    : Document location			*/
/*  (returns)	: KBDocRoot	    :					*/

KBDocRoot::KBDocRoot
	(	KBNode			*parent,
		QPtrList<KBNode>	&children,
		const KBLocation	&location
	)
	:
	m_parent	(parent),
	m_children	(children),
	m_location 	(location)
{
	m_progress	= 0	;
	m_locking	= 0	;
	m_loading	= false	;
	m_mimeSource	= 0	;
	m_serverInfo	= m_location.getServerInfo() ;
	reset () ;
}

/*  KBDocRoot								*/
/*  ~KBDoctRoot	: Common document root class destructor			*/

KBDocRoot::~KBDocRoot ()
{
	DELOBJ	(m_paramDict )	;
	DELOBJ	(m_mimeSource)	;
}

KBMimeSourceFactory
	*KBDocRoot::mimeSource ()
{
	if (m_mimeSource == 0) m_mimeSource = new KBMimeSourceFactory(this) ;
	return	m_mimeSource ;
}

/*  KBDocRoot								*/
/*  setParamDict: Set parameter dictionary				*/
/*  _paramDict	: QDict<QString>&: Parameter dictionary			*/
/*  pError	: KBError &	 : Error return				*/
/*  (returns)	: KB::ShowRC	 : OK/Error/Cancel			*/

KB::ShowRC
	KBDocRoot::setParamDict
	(	const QDict<QString>	&_paramDict,
		KBError			&pError
	)
{
	/* Delete any existing parameter dictionary and set a new copy	*/
	/* of the new one. Then scan the document to locate all		*/
	/* defined paramaters; if there are none then we can return	*/
	/* without anything more to do.					*/
	DELOBJ	(m_paramDict) ;
	m_paramDict = new KBAttrDict (_paramDict) ;

	QDict<KBParamSet> paramSet    ;
	paramSet.setAutoDelete  (true) ;
	m_parent->findAllParams (paramSet) ;

	if (paramSet.count() == 0)
		return KB::ShowRCOK ;

	/* Scan the parameter dictionary and update any matching	*/
	/* parameters with the supplied values. Hence passed parameter	*/
	/* values override parameters defined in the document.		*/
	QDictIterator<QString> pIter (*m_paramDict) ;
	while (pIter.current() != 0)
	{
		KBParamSet *pSet = paramSet.find(pIter.currentKey()) ;
		if (pSet != 0)
		{
			pSet->m_value	= *pIter.current() ;
			pSet->m_set	= true	;
		}
		pIter += 1 ;
	}

	/* Run the user paramater set dialog, unless there we no	*/
	/* parameters left to set. If the user cancels this then	*/
	/* return in error.						*/
	bool	      ok    ;
	KBParamSetDlg pDlg  (TR("Set Parameters"), paramSet, this, pError, ok) ;
	if (!ok) return	 KB::ShowRCError ;

	if (pDlg.anyParams() && !pDlg.exec ())
	{
		pError	= KBError
			  (	KBError::Warning,
				TR("User cancelled parameter dialog"),
				QString::null,
				__ERRLOCN
			  )	;
		return	KB::ShowRCCancel ;
	}

	QDictIterator<KBParamSet> iter (paramSet) ;
	KBParamSet		  *pSet		  ;

	while ((pSet = iter.current()) != 0)
	{
		m_paramDict->replace (iter.currentKey(), new QString(pSet->m_value)) ;
		iter += 1 ;
	}

	return	KB::ShowRCOK ;
}

/*  KBDocRoot								*/
/*  reset	: Reset document root information			*/
/*  (returns)	: void		 :					*/

void	KBDocRoot::reset ()
{
	m_scrIface[0]  		= 0     ;
	m_scrIface[1] 		= 0     ;
	m_paramDict 		= 0	;
	m_gotExecError		= false	;
}

/*  KBDocRoot								*/
/*  doExecError	: Execution error reported				*/
/*  (returns)	: void		:					*/

void	KBDocRoot::doExecError ()
{
	if (!m_gotExecError)
	{
		m_gotExecError	= true	;
		emit	execError ()	;
	}
}

/*  KBDocRoot								*/
/*  doRequestClose							*/
/*		: Request to close the document				*/
/*  rc		: int		: Return code				*/
/*  (returns)	: void		:					*/

void	KBDocRoot::doRequestClose
	(	int	rc
	)
{
	emit	requestClose (rc) ;
}

/*  KBDocRoot								*/
/*  doSetMessage: Display message in status bar				*/
/*  message	: const QString &    : Message				*/
/*  (returns)	: void		     :					*/

void	KBDocRoot::doSetMessage
	(	const QString    	&message
	)
{
	if (m_message)
		m_message->setText (message) ;
}

/*  KBDocRoot								*/
/*  doSetLocking: Display locking state in status bar			*/
/*  locking	: KBQryBase::Locking : Locking state			*/
/*  (returns)	: void		     :					*/

void	KBDocRoot::doSetLocking
	(	KBQryBase::Locking	locking
	)
{
	if (m_locking)
		m_locking->setText
		(	locking == KBQryBase::NoLocking ?
				TR("Unlocked") :
				TR("Record locked")
		)	;
}

/*  KBDocRoot								*/
/*  getParamValue: Get value of named parameter				*/
/*  param	 : cchar *	: Parameter				*/
/*  (returns)	 : QString &	: Parameter value			*/

QString	KBDocRoot::getParamValue
	(	cchar	*param
	)
{
	/* The parameter may be either "name" or "name:defval", so we	*/
	/* split on the colon first ...					*/
	QString		res	= QString::null ;
	QStringList	bits	= QStringList::split (':', param) ;

	/* ... then, if there is a paramater dictionary, see if the	*/
	/* parameter has been defined.					*/
	if (m_paramDict != 0)
	{	QString	*val = m_paramDict->find (bits[0]) ;
		if (val != 0) res = *val ;
	}

	/* If we don't have a value, and the parameter was of the form	*/
	/* "name:defval", then use that default value.			*/
	if (res.isEmpty() && (bits.count() > 1)) res = bits[1] ;

	return	res	;
}

/*  KBDocRoot								*/
/*  setParamValue: Set value of named parameter				*/
/*  name	 : cchar *	   : Parameter name			*/
/*  value	 : const QString & : Parameter value			*/
/*  (returns)	 : void		   :					*/

void	KBDocRoot::setParamValue
	(	cchar		*name,
		const QString	&value
	)
{
	if (m_paramDict != 0)
		m_paramDict->replace (name, new QString(value)) ;
}

/*  KBDocRoot								*/
/*  getScriptIF	: Get script interface					*/
/*  l2		: bool		  : Use second language			*/
/*  ok		: bool &	  : Error flag				*/
/*  pError	: KBError &	  : Error return			*/
/*  (returns)	: KBScriptIF *	  : Script interface or null on error	*/

KBScriptIF
	*KBDocRoot::getScriptIF
	(	bool		l2,
		bool		&ok,
		KBError		&pError
	)
{
	/* This method is used to get a scripting interface in order	*/
	/* to verify a script (in an event or a slot). It (a) does	*/
	/* assume that there is a defined language and (b) does not do	*/
	/* and script module loading or import setup/			*/
	KBScriptIF	*iface	= 0 ;
	QString		language    ;

	if (l2)
		language = m_parent->getAttrVal ("language2") ;
	else	language = m_parent->getAttrVal ("language" ) ;

	fprintf
	(	stderr,
		"KBDocRoot::getScriptIF: l2=%d l=%s\n",
		l2,
		(cchar *)language
	)	;

	if (language.isEmpty())
	{
		pError	= KBError
			  (	KBError::Error,
				TR("No scripting language specified"),
				QString::null,
				__ERRLOCN
			  )	;
		ok	= false	;
		return	0	;
	}

	if ((iface = LinkKBScript (language, pError)) == 0)
	{	ok	= false ;
		return	0	;
	}

	ok	= true	;
	return	iface	;
}

/*  KBDocRoot								*/
/*  loadScripting: Load scripting and modules				*/
/*  l2		: bool		  : Use second language			*/
/*  ok		 : bool &	  : Error flag				*/
/*  pError	 : KBError &	  : Error return			*/
/*  (returns)	 : KBScriptIF *	  : Script interface or null on error	*/

KBScriptIF
	*KBDocRoot::loadScripting
	(	bool		l2,
		bool		&ok,
		KBError		&pError
	)
{
	/* See if a scripting interface is already present, in which	*/
	/* case return that.						*/
	int	slot	= l2 ? 1 : 0 ;

	if (m_scrIface[slot] != 0)
	{	ok	= true	;
		return	m_scrIface[slot] ;
	}

	/* See if a script language is specified, in which case we need	*/
	/* to load its interface. Set an error and return failure if	*/
	/* this fails.							*/
	QString	 language = m_parent->getAttrVal ("language") ;
	if (!language.isEmpty())
		if ((m_scrIface[0] = LinkKBScript (language, pError)) == 0)
		{	ok	= false ;
			return	0	;
		}

	/* See if a second script language is specified, in which case	*/
	/* we need to load its interface. Set an error and return	*/
	/* failure if this fails.					*/
	QString	 language2 = m_parent->getAttrVal ("language2") ;
	if (!language2.isEmpty())
		if ((m_scrIface[1] = LinkKBScript (language2, pError)) == 0)
		{	ok	= false ;
			return	0	;
		}




	m_scripts.clear () ;
	m_imports.clear () ;

	LITER
	(	KBNode,
		m_children,
		child,

		if (child->isImport() != 0)
			m_imports.append (child->isImport()->getModuleName()) ;
		if (child->isScript() != 0)
			m_scripts.append (child->isScript()->getModuleName()) ;
	)

	/* Load all specified script modules. We check inside the loop	*/
	/* that we have a scripting interface, so that the user need	*/
	/* not set a language unless there is a script.			*/
	LITER
	(	KBNode,
		m_children,
		node,

		KBScript	*script ;
		if ((script = node->isScript()) == 0) continue ;

		QString		name	 = script->getModuleName() ;
		KBScriptIF	*scrIF	 = m_scrIface[0] ;
		QString		lang	 = language	 ;

		if (name.left(2) == "2.")
		{
			name	= name.mid  (2) ;
			scrIF	= m_scrIface[1] ;
			lang	= language2	;
		}

		if (scrIF == 0)
		{
			pError	= KBError
				  (	KBError::Error,
					"No script language",
					QString("Modules %1,...").arg(name),
					__ERRLOCN
				  )	;
			ok	= false	;
			return	0	;
		}

		KBLocation scrloc
			   (	m_location.dbInfo(),
				"script",
				m_location.server(),
				name,
				lang
			   )	  ;
		QString	   eText  ;
		QString	   ePatt  ;

		if (!scrIF->load (scrloc, eText, ePatt, pError))
		{
			fprintf
			(	stderr,
				"KBDocRoot::loadScripting: scrIface->load failed [%s][%s][%s][%s]\n",
				(cchar *)scrloc.ident(),
				(cchar *)eText,
				(cchar *)pError.getMessage(),
				(cchar *)pError.getDetails()
			)	;

			KBAppPtr::getCallback()->editScript (scrloc, eText, ePatt, 0, pError) ;
			ok	= false	;
			return	0	;
		}
	)


	ok	= true	   ;
	return	m_scrIface[slot] ;
}

/*  KBDocRoot								*/
/*  doLayoutChanged							*/
/*		: Notify layout has been changed			*/
/*  (returns)	: void		:					*/

void	KBDocRoot::doLayoutChanged ()
{
//	KBForm	*form	= m_parent->isForm() ;
//	if (form != 0)
//	{
//		QSize	size	= form->sizeHint() ;
//		form->resize (form->sizeHint()) ;
//
//		fprintf
//		(	stderr,
//			"KBDocRoot::doLayoutChanged: %s\n",
//			_TEXT(size)
//		)	;
//	}
//
	emit	layoutChanged () ;
}

void	KBDocRoot::setStatusBar
	(	QLabel		*message,
		QLabel		*locking,
		KBProgressBox	*progress
	)
{
	m_message  = message	;
	m_locking  = locking	;
	m_progress = progress	;

	if (m_locking)
		m_locking->setText (TR("Unlocked")) ;

}

void	KBDocRoot::loadingStart ()
{
	if (m_progress) m_progress->start() ;
	m_loading = true  ;
}

void	KBDocRoot::loadingDone ()
{
	m_loading = false ;
	if (m_progress) m_progress->stop() ;
}


/*  KBDocRoot								*/
/*  loadingProgress							*/
/*		: Set progress						*/
/*  total	: uint		: Target value				*/
/*  done	: uint		: Done so far				*/
/*  message	: QString	: Message				*/
/*  (returns)	: bool		: True if cancelled			*/

bool	KBDocRoot::loadingProgress
	(	uint		total,
		uint		done,
		QString		message
	)
{
//	fprintf
//	(	stderr,
//		"KBDocRoot::setProgress: %d %d [%s]\n",
//		total,
//		done,
//		(cchar *)message
//	)	;

	if (m_progress)
	{
		m_progress->setTotal (total)	;
		m_progress->setDone  (done )	;
		return	m_progress->cancelled() ;
	}

	return	false	;
}

/*  KBDocRoot								*/
/*  appFont	: Get application font for the document			*/
/*  (returns)	: const QString & : Font specification			*/

const QString
	&KBDocRoot::appFont ()
{
	/* Check for serverInfo, since this will not be set if the form	*/
	/* is running as part of the test suite.			*/
	if (m_serverInfo == 0)
	{
		static	QString	dummy	;
		return	dummy		;
	}

	return	m_serverInfo->appFont()	;
}

/*  KBDocRoot								*/
/*  gotExecError: Get reference to error flag				*/
/*  (returns)	: const bool &	: Reference to error flag		*/

const bool
	&KBDocRoot::gotExecError ()
{
	return	m_gotExecError	;
}
