/***************************************************************************
    file	         : kb_pyvalue.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
			   (C)	2005
			   	Abel Deuring
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include "kb_pyvalue.h"
#include "kb_pyscript.h"
#include "kb_pybase.h"
#include "structmember.h"

#include "kb_type.h"
#include "kb_value.h"

#include "kb_options.h"



static	QTextCodec *__pyCodec = 0 ;

/*  makeCodec	: Return coded for current python encoding		*/
/*  (returns)	: QTextCodec *	: Codec					*/

static	QTextCodec *makeCodec()
{
	/* Codec is kept over calls, only make it when there is not one	*/
	/* already.							*/
	if (__pyCodec == 0)
		__pyCodec = QTextCodec::codecForName(KBOptions::getPythonEncoding());
	return	__pyCodec ;
}

/*  resetCodec	: Reset (ie., clear) codec				*/
/*  (returns)	: void		:					*/

void	resetCodec ()
{
	/* Just clear the codec pointer. A new one will be set up next	*/
	/* time "makeCodec" is called.					*/
	__pyCodec = 0 ;
}


/* kb_qStringToPyString							*/
/*		: convert a QString into a Python string		*/
/* qs		: QString to be converted				*/
/* (returns)	: Python string object 					*/
/* xxx allow explicit conversion into a Python Unicode object??		*/
PyObject *kb_qStringToPyString
	(const QString& qs)
{
	QTextCodec *codec = makeCodec() ;
	if (codec == 0)
		// xxx a bit too simplistic: 
		return PyErr_NoMemory();
	
	return PyString_FromString(codec->fromUnicode(qs));
}


/* kb_pyStringToQString							*/
/*		: convert a Python String into a QString		*/
/* po		: Python object to be converted				*/
/* error	: on return true, if a Python exception occurred	*/
/*		: the caller must clear or process the exception	*/
/* (returns)	: Python string object 					*/
QString kb_pyStringToQString
	(PyObject *po, bool &error)
{
	error = false;
	if (PyUnicode_Check(po))
	{
		PyObject *pUtf8 = PyUnicode_EncodeUTF8
					(PyUnicode_AS_UNICODE(po),
					 PyUnicode_GET_SIZE(po),
					 "strict"
					);
		if (pUtf8 == 0)
		{
			error = true;
			return "";
		}
		
		QString res = QString::fromUtf8(PyString_AsString(pUtf8));
		Py_DECREF(pUtf8);
		return res;
	}
	
	if (PyString_Check(po))
	{
		QTextCodec *codec = makeCodec() ;
		if (codec == 0)
		{
			error = true;
			return "";
		}
		return codec->toUnicode(PyString_AsString(po));
	}
	
	// try to convert the object into a string
	PyObject *asStr = PyObject_Str(po);
	if (asStr == 0)
	{
		error = true;
		return "";
	}
	
	QString res = kb_pyStringToQString(asStr, error);
	Py_DECREF(asStr);
	return res;
}

QCString pyEncodeSourcecode(QString s, bool &error)
{
	QTextCodec *codec = makeCodec() ;
	if (codec == 0)
	{
		error = true;
		PyErr_NoMemory();
		return "";
	}
	
	error = false;
	return codec->fromUnicode(s);
}

/* PyKBValue: Python class that represents/contains a KBValue object 	*/

#define FIXED		0
#define FLOAT		1
#define	DATE		2
#define	TIME		3
#define	DATETIME	4
#define STRING		5
#define	BINARY		6
#define	BOOL		7


#define TYPE_ENTRY(x) {x, #x}
static struct
{
	int		num;
	const char*	name;
} kbtypes[] =
{
	TYPE_ENTRY(FIXED),
	TYPE_ENTRY(FLOAT),
	TYPE_ENTRY(DATE),
	TYPE_ENTRY(TIME),
	TYPE_ENTRY(DATETIME),
	TYPE_ENTRY(STRING),
	TYPE_ENTRY(BINARY),
	TYPE_ENTRY(BOOL),
	{-1, 	0}
};

typedef struct
{
	PyObject_HEAD
	KBValue	*value;
} pyKBValue;


static PyObject *PyKBValue_new(PyTypeObject *type, PyObject *args, PyObject *kw)
{
	pyKBValue *self;
	self = (pyKBValue*)type->tp_alloc(type, 0);
	if (self == 0)
		return 0;
	
	self->value = new KBValue();
	
	return (PyObject*) self;
}

static int PyKBValue_init(pyKBValue *self, PyObject *args, PyObject *kw)
{
	PyObject 	*pyVal, *asStr;
	int 		type = STRING;
	long		ival;
	double		dval;
	QString		sval;
	QByteArray	bval;
	bool		error;
	
	if (!PyArg_ParseTuple(args, "O|i", &pyVal, &type))
		return -1;
	
	switch (type)
	{
		case	FIXED:
			ival = PyInt_AsLong(pyVal);
			if (PyErr_Occurred())
				return -1;
			delete self->value;
			self->value = new KBValue(ival, &_kbFixed);
			break;

		case	FLOAT:
			dval = PyFloat_AsDouble(pyVal);
			if (PyErr_Occurred())
				return -1;
			delete self->value;
			self->value = new KBValue(dval, &_kbFloat);
			break;

		case	DATE:
			// xxx remember: Python 2.4 has a special DateTime objects
			PyErr_SetString(PyKBRekallError, 
				"conversion to KB::ITDate type not yet implemented");
			return -1;
		
		case	TIME:
			// xxx remember: Python 2.4 has a special DateTime objects
			PyErr_SetString(PyKBRekallError, 
				"conversion to KB::ITTime type not yet implemented");
			return -1;
		
		case	DATETIME:
			// xxx remember: Python 2.4 has a special DateTime objects
			PyErr_SetString(PyKBRekallError, 
				"conversion to KB::ITDateTime type not yet implemented");
			return -1;
		
		case	STRING:
			sval = kb_pyStringToQString(pyVal, error);
			if (error)
				return -1;
			delete self->value;
			self->value = new KBValue(sval, &_kbString);
			break;
		
		case	BINARY:
			asStr = PyObject_Str(pyVal);
			if (asStr == 0)
				return -1;
			bval.duplicate(PyString_AS_STRING(asStr), 
					PyString_Size(asStr));
			delete self->value;
			self->value = new KBValue(bval, &_kbBinary);
			break;
			
		case	BOOL:
			delete self->value;
			self->value = new KBValue(PyObject_IsTrue(pyVal),
						  &_kbBool);
			break;
			
		default:
			PyErr_SetString(PyExc_TypeError, "invalid KB type");
			return -1;
	}
	if (self->value == 0)
	{
		PyErr_NoMemory();
		return -1;
	}
	return 0;
}

static void PyKBValue_dealloc(pyKBValue *self)
{
	delete self->value;
	self->ob_type->tp_free((PyObject*) self);
}

static PyMethodDef PyKBValue_methods[] =
{
	{0}
};

static PyMemberDef PyKBValue_members[] = 
{
	{0}
};

PyTypeObject pyKBValueType =
{
	PyObject_HEAD_INIT(NULL)
	0,				/* ob_size 		*/
	"RekallMain.KBValue",		/* tp_name 		*/
	sizeof(pyKBValue),		/* tp_basicsize		*/
	0,				/* tp_itemsize		*/
	(destructor) PyKBValue_dealloc,	/* tp_dealloc		*/
	0,				/* tp_print		*/
	0,				/* tp_getattr		*/
	0,				/* tp_setattr		*/
	0,				/*tp_compare		*/
	0,				/*tp_repr		*/
	0,				/*tp_as_number		*/
	0,				/*tp_as_sequence	*/
	0,				/*tp_as_mapping		*/
	0,				/*tp_hash 		*/
	0,				/*tp_call		*/
	0,				/*tp_str		*/
	0,				/*tp_getattro		*/
	0,				/*tp_setattro		*/
	0,				/*tp_as_buffer		*/
	Py_TPFLAGS_DEFAULT,		/*tp_flags		*/
	"xxx no docs yet",		/* tp_doc		*/
	0,				/* tp_traverse		*/
	0,				/* tp_clear 		*/
	0,				/* tp_richcompare 	*/
	0,				/* tp_weaklistoffset 	*/
	0,				/* tp_iter 		*/
	0,				/* tp_iternext 		*/
	PyKBValue_methods,		/* tp_methods 		*/
	PyKBValue_members,		/* tp_members 		*/
	0,				/* tp_getset 		*/
	0,				/* tp_base 		*/
	0,				/* tp_dict 		*/
	0,				/* tp_descr_get 	*/
	0,				/* tp_descr_set 	*/
	0,				/* tp_dictoffset 	*/
	(initproc)PyKBValue_init,	/* tp_init 		*/
	0,				/* tp_alloc 		*/
	PyKBValue_new,			/* tp_new 		*/

};

KBValue *kb_pyKBValueCheck(PyObject *po)
{
	if (po->ob_type == &pyKBValueType)
	{
		return ((pyKBValue*)po)->value;
	}
	else	return 0;
}


void	initPyValue(PyObject *module)
{
	if (PyType_Ready(&pyKBValueType) < 0)
		return;
	
	int i = 0;
	while (kbtypes[i].num >= 0)
	{
		fprintf(stderr, "type index %i\n", i);
		PyObject	*val = PyInt_FromLong(kbtypes[i].num);
		if (-1 == PyDict_SetItemString 
				(
					pyKBValueType.tp_dict,
					kbtypes[i].name,
					val
				))
			return; // need to signal an error!!!
		Py_DECREF(val);
		i++;
	}
	
	Py_INCREF(&pyKBValueType);
	PyModule_AddObject(module, "KBValue", (PyObject*) &pyKBValueType);
	
}
